@@ -19,52 +19,47 @@ import logging
 from django.conf import settings
 from django.contrib import auth
 from django.contrib.auth.decorators import login_required
-try:
-    from django.contrib.auth.views import LogoutView
-    django_logout = LogoutView.as_view()
-except ImportError:
-    from django.contrib.auth.views import logout as django_logout
 from django.core.exceptions import PermissionDenied, SuspiciousOperation
-from django.http import Http404, HttpResponse
-from django.http import HttpResponseRedirect  # 30x
 from django.http import HttpResponseBadRequest  # 40x
+from django.http import HttpResponseRedirect  # 30x
 from django.http import HttpResponseServerError  # 50x
-from django.views.decorators.http import require_POST
+from django.http import Http404, HttpResponse
 from django.shortcuts import render
 from django.template import TemplateDoesNotExist
-
-try:
-    from django.utils.six import text_type, binary_type, PY3
-except ImportError:
-    import sys
-    PY3 = sys.version_info[0] == 3
-    text_type = str
-    binary_type = bytes
-
 from django.views.decorators.csrf import csrf_exempt
+from django.views.generic import View
+from django.utils.decorators import method_decorator
 
 from saml2 import BINDING_HTTP_REDIRECT, BINDING_HTTP_POST
+from saml2.client_base import LogoutError
 from saml2.metadata import entity_descriptor
 from saml2.ident import code, decode
-from saml2.sigver import MissingKey
 from saml2.s_utils import UnsupportedBinding
 from saml2.response import (
     StatusError, StatusAuthnFailed, SignatureError, StatusRequestDenied,
     UnsolicitedResponse, StatusNoAuthnContext,
 )
+from saml2.mdstore import SourceNotFound
+from saml2.sigver import MissingKey
+from saml2.samlp import AuthnRequest
 from saml2.validate import ResponseLifetimeExceed, ToEarly
-from saml2.xmldsig import SIG_RSA_SHA1, SIG_RSA_SHA256  # support for SHA1 is required by spec
+from saml2.xmldsig import (  # support for SHA1 is required by spec
+    SIG_RSA_SHA1, SIG_RSA_SHA256)
 
-from djangosaml2.cache import IdentityCache, OutstandingQueriesCache
-from djangosaml2.cache import StateCache
-from djangosaml2.exceptions import IdPConfigurationMissing
-from djangosaml2.conf import get_config
-from djangosaml2.overrides import Saml2Client
-from djangosaml2.signals import post_authenticated
-from djangosaml2.utils import (
-    available_idps, fail_acs_response, get_custom_setting,
-    get_idp_sso_supported_bindings, get_location, is_safe_url_compat,
-)
+from .cache import IdentityCache, OutstandingQueriesCache, StateCache
+from .conf import get_config
+from .exceptions import IdPConfigurationMissing
+from .overrides import Saml2Client
+from .signals import post_authenticated
+from .utils import (available_idps, fail_acs_response, get_custom_setting,
+                    get_idp_sso_supported_bindings, get_location,
+                    validate_referral_url)
+
+try:
+    from django.contrib.auth.views import LogoutView
+    django_logout = LogoutView.as_view()
+except ImportError:
+    from django.contrib.auth.views import logout as django_logout
 
 
 logger = logging.getLogger('djangosaml2')
@@ -81,16 +76,6 @@ def _get_subject_id(session):
         return None
 
 
-def callable_bool(value):
-    """ A compatibility wrapper for pre Django 1.10 User model API that used
-    is_authenticated() and is_anonymous() methods instead of attributes
-    """
-    if callable(value):
-        return value()
-    else:
-        return value
-
-
 def login(request,
           config_loader_path=None,
           wayf_template='djangosaml2/wayf.html',
@@ -117,10 +102,7 @@ def login(request,
     if not came_from:
         logger.warning('The next parameter exists but is empty')
         came_from = settings.LOGIN_REDIRECT_URL
-
-    # Ensure the user-originating redirection url is safe.
-    if not is_safe_url_compat(url=came_from, allowed_hosts={request.get_host()}):
-        came_from = settings.LOGIN_REDIRECT_URL
+    came_from = validate_referral_url(request, came_from)
 
     # if the user is already authenticated that maybe because of two reasons:
     # A) He has this URL in two browser windows and in the other one he
@@ -132,7 +114,7 @@ def login(request,
     # SAML_IGNORE_AUTHENTICATED_USERS_ON_LOGIN setting. If that setting
     # is True (default value) we will redirect him to the came_from view.
     # Otherwise, we will show an (configurable) authorization error.
-    if callable_bool(request.user.is_authenticated):
+    if request.user.is_authenticated:
         redirect_authenticated_user = getattr(settings, 'SAML_IGNORE_AUTHENTICATED_USERS_ON_LOGIN', True)
         if redirect_authenticated_user:
             return HttpResponseRedirect(came_from)
@@ -143,7 +125,15 @@ def login(request,
                     })
 
     selected_idp = request.GET.get('idp', None)
-    conf = get_config(config_loader_path, request)
+    try:
+        conf = get_config(config_loader_path, request)
+    except SourceNotFound as excp:
+        msg = ('Error, IdP EntityID was not found '
+               'in metadata: {}')
+        logger.exception(msg.format(excp))
+        return HttpResponse(msg.format(('Please contact '
+                                        'technical support.')),
+                            status=500)
 
     kwargs = {}
     # pysaml needs a string otherwise: "cannot serialize True (type bool)"
@@ -167,7 +157,7 @@ def login(request,
                                            'its metadata is expired.'))
         if selected_idp is None:
             selected_idp = list(idps.keys())[0]
-    
+
     # choose a binding to try first
     sign_requests = getattr(conf, '_sp_authn_requests_signed', False)
     binding = BINDING_HTTP_POST if sign_requests else BINDING_HTTP_REDIRECT
@@ -197,20 +187,21 @@ def login(request,
     logger.debug('Redirecting user to the IdP via %s binding.', binding)
     if binding == BINDING_HTTP_REDIRECT:
         try:
-            # do not sign the xml itself, instead use the sigalg to
-            # generate the signature as a URL param
-            sig_alg_option_map = {'sha1': SIG_RSA_SHA1,
-                                  'sha256': SIG_RSA_SHA256}
-            sig_alg_option = getattr(conf, '_sp_authn_requests_signed_alg', 'sha1')
-            sigalg = sig_alg_option_map[sig_alg_option] if sign_requests else None
             nsprefix = get_namespace_prefixes()
+            if sign_requests:
+                # do not sign the xml itself, instead use the sigalg to
+                # generate the signature as a URL param
+                sig_alg_option_map = {'sha1': SIG_RSA_SHA1,
+                                      'sha256': SIG_RSA_SHA256}
+                sig_alg_option = getattr(conf, '_sp_authn_requests_signed_alg', 'sha1')
+                kwargs["sigalg"] = sig_alg_option_map[sig_alg_option]
             session_id, result = client.prepare_for_authenticate(
                 entityid=selected_idp, relay_state=came_from,
-                binding=binding, sign=False, sigalg=sigalg,
-                nsprefix=nsprefix, **kwargs)
+                binding=binding, sign=False, nsprefix=nsprefix,
+                **kwargs)
         except TypeError as e:
             logger.error('Unable to know which IdP to use')
-            return HttpResponse(text_type(e))
+            return HttpResponse(str(e))
         else:
             http_response = HttpResponseRedirect(get_location(result))
     elif binding == BINDING_HTTP_POST:
@@ -220,16 +211,16 @@ def login(request,
                 location = client.sso_location(selected_idp, binding)
             except TypeError as e:
                 logger.error('Unable to know which IdP to use')
-                return HttpResponse(text_type(e))
+                return HttpResponse(str(e))
             session_id, request_xml = client.create_authn_request(
                 location,
                 binding=binding,
                 **kwargs)
             try:
-                if PY3:
-                    saml_request = base64.b64encode(binary_type(request_xml, 'UTF-8')).decode('utf-8')
-                else:
-                    saml_request = base64.b64encode(binary_type(request_xml))
+                if isinstance(request_xml, AuthnRequest):
+                    # request_xml will be an instance of AuthnRequest if the message is not signed
+                    request_xml = str(request_xml)
+                saml_request = base64.b64encode(bytes(request_xml, 'UTF-8')).decode('utf-8')
 
                 http_response = render(request, post_binding_form_template, {
                     'target_url': location,
@@ -249,117 +240,175 @@ def login(request,
                     binding=binding)
             except TypeError as e:
                 logger.error('Unable to know which IdP to use')
-                return HttpResponse(text_type(e))
+                return HttpResponse(str(e))
             else:
                 http_response = HttpResponse(result['data'])
     else:
         raise UnsupportedBinding('Unsupported binding: %s', binding)
 
     # success, so save the session ID and return our response
-    logger.debug('Saving the session_id in the OutstandingQueries cache')
-    oq_cache = OutstandingQueriesCache(request.session)
+    oq_cache = OutstandingQueriesCache(request.saml_session)
     oq_cache.set(session_id, came_from)
+    logger.debug('Saving the session_id "{}" in the OutstandingQueries cache'.format(oq_cache.__dict__))
     return http_response
 
 
-@require_POST
-@csrf_exempt
-def assertion_consumer_service(request,
-                               config_loader_path=None,
-                               attribute_mapping=None,
-                               create_unknown_user=None):
-    """SAML Authorization Response endpoint
-
-    The IdP will send its response to this view, which
-    will process it with pysaml2 help and log the user
-    in using the custom Authorization backend
-    djangosaml2.backends.Saml2Backend that should be
-    enabled in the settings.py
+class AssertionConsumerServiceView(View):
+    """
+    The IdP will send its response to this view, which will process it using pysaml2 and
+    log the user in using whatever SAML authentication backend has been enabled in
+    settings.py. The `djangosaml2.backends.Saml2Backend` can be used for this purpose,
+    though some implementations may instead register their own subclasses of Saml2Backend.
     """
-    attribute_mapping = attribute_mapping or get_custom_setting('SAML_ATTRIBUTE_MAPPING', {'uid': ('username', )})
-    create_unknown_user = create_unknown_user if create_unknown_user is not None else \
-                          get_custom_setting('SAML_CREATE_UNKNOWN_USER', True)
-    conf = get_config(config_loader_path, request)
-    try:
-        xmlstr = request.POST['SAMLResponse']
-    except KeyError:
-        logger.warning('Missing "SAMLResponse" parameter in POST data.')
-        raise SuspiciousOperation
 
-    client = Saml2Client(conf, identity_cache=IdentityCache(request.session))
+    @method_decorator(csrf_exempt)
+    def dispatch(self, request, *args, **kwargs):
+        """
+        This view needs to be CSRF exempt because it is called prior to login.
+        """
+        return super(AssertionConsumerServiceView, self).dispatch(request, *args, **kwargs)
 
-    oq_cache = OutstandingQueriesCache(request.session)
-    outstanding_queries = oq_cache.outstanding_queries()
+    @method_decorator(csrf_exempt)
+    def post(self,
+             request,
+             config_loader_path=None,
+             attribute_mapping=None,
+             create_unknown_user=None):
+        """
+        SAML Authorization Response endpoint
+        """
+        attribute_mapping = attribute_mapping or \
+                            get_custom_setting('SAML_ATTRIBUTE_MAPPING',
+                                               {'uid': ('username', )})
+        create_unknown_user = create_unknown_user or \
+                              get_custom_setting('SAML_CREATE_UNKNOWN_USER', True)
+        conf = get_config(config_loader_path, request)
+        try:
+            xmlstr = request.POST['SAMLResponse']
+        except KeyError:
+            logger.warning('Missing "SAMLResponse" parameter in POST data.')
+            raise SuspiciousOperation
 
-    try:
-        response = client.parse_authn_request_response(xmlstr, BINDING_HTTP_POST, outstanding_queries)
-    except (StatusError, ToEarly):
-        logger.exception("Error processing SAML Assertion.")
-        return fail_acs_response(request)
-    except ResponseLifetimeExceed:
-        logger.info("SAML Assertion is no longer valid. Possibly caused by network delay or replay attack.", exc_info=True)
-        return fail_acs_response(request)
-    except SignatureError:
-        logger.info("Invalid or malformed SAML Assertion.", exc_info=True)
-        return fail_acs_response(request)
-    except StatusAuthnFailed:
-        logger.info("Authentication denied for user by IdP.", exc_info=True)
-        return fail_acs_response(request)
-    except StatusRequestDenied:
-        logger.warning("Authentication interrupted at IdP.", exc_info=True)
-        return fail_acs_response(request)
-    except StatusNoAuthnContext:
-        logger.warning("Missing Authentication Context from IdP.", exc_info=True)
-        return fail_acs_response(request)
-    except MissingKey:
-        logger.exception("SAML Identity Provider is not configured correctly: certificate key is missing!")
-        return fail_acs_response(request)
-    except UnsolicitedResponse:
-        logger.exception("Received SAMLResponse when no request has been made.")
-        return fail_acs_response(request)
+        client = Saml2Client(conf, identity_cache=IdentityCache(request.saml_session))
+        oq_cache = OutstandingQueriesCache(request.saml_session)
+        oq_cache.sync()
+        outstanding_queries = oq_cache.outstanding_queries()
 
-    if response is None:
-        logger.warning("Invalid SAML Assertion received (unknown error).")
-        return fail_acs_response(request, status=400, exc_class=SuspiciousOperation)
+        _exception = None
+        try:
+            response = client.parse_authn_request_response(xmlstr,
+                                                           BINDING_HTTP_POST,
+                                                           outstanding_queries)
+        except (StatusError, ToEarly) as e:
+            _exception = e
+            logger.exception("Error processing SAML Assertion.")
+        except ResponseLifetimeExceed as e:
+            _exception = e
+            logger.info(("SAML Assertion is no longer valid. "
+                         "Possibly caused by network delay or replay attack."),
+                         exc_info=True)
+        except SignatureError as e:
+            _exception = e
+            logger.info("Invalid or malformed SAML Assertion.", exc_info=True)
+        except StatusAuthnFailed as e:
+            _exception = e
+            logger.info("Authentication denied for user by IdP.", exc_info=True)
+        except StatusRequestDenied as e:
+            _exception = e
+            logger.warning("Authentication interrupted at IdP.", exc_info=True)
+        except StatusNoAuthnContext as e:
+            _exception = e
+            logger.warning("Missing Authentication Context from IdP.", exc_info=True)
+        except MissingKey as e:
+            _exception = e
+            logger.exception("SAML Identity Provider is not configured correctly: "
+                             "certificate key is missing!")
+        except UnsolicitedResponse as e:
+            _exception = e
+            logger.exception("Received SAMLResponse when no request has been made.")
 
-    session_id = response.session_id()
-    oq_cache.delete(session_id)
 
-    # authenticate the remote user
-    session_info = response.session_info()
+        if _exception:
+            return fail_acs_response(request, exception=_exception)
+        elif response is None:
+            logger.warning("Invalid SAML Assertion received (unknown error).")
+            return fail_acs_response(request, status=400,
+                                     exception=SuspiciousOperation('Unknown SAML2 error'))
 
-    if callable(attribute_mapping):
-        attribute_mapping = attribute_mapping()
-    if callable(create_unknown_user):
-        create_unknown_user = create_unknown_user()
+        session_id = response.session_id()
+        oq_cache.delete(session_id)
 
-    logger.debug('Trying to authenticate the user. Session info: %s', session_info)
-    user = auth.authenticate(request=request,
-                             session_info=session_info,
-                             attribute_mapping=attribute_mapping,
-                             create_unknown_user=create_unknown_user)
-    if user is None:
-        logger.warning("Could not authenticate user received in SAML Assertion. Session info: %s", session_info)
-        raise PermissionDenied
+        # authenticate the remote user
+        session_info = response.session_info()
 
-    auth.login(request, user)
-    _set_subject_id(request.session, session_info['name_id'])
-    logger.debug("User %s authenticated via SSO.", user)
+        if callable(attribute_mapping):
+            attribute_mapping = attribute_mapping()
+        if callable(create_unknown_user):
+            create_unknown_user = create_unknown_user()
 
-    logger.debug('Sending the post_authenticated signal')
-    post_authenticated.send_robust(sender=user, session_info=session_info)
+        logger.debug('Trying to authenticate the user. Session info: %s', session_info)
+        user = auth.authenticate(request=request,
+                                 session_info=session_info,
+                                 attribute_mapping=attribute_mapping,
+                                 create_unknown_user=create_unknown_user)
+        if user is None:
+            logger.warning("Could not authenticate user received in SAML Assertion. Session info: %s", session_info)
+            return fail_acs_response(request, exception=PermissionDenied('No user could be authenticated.'))
+
+        auth.login(self.request, user)
+        _set_subject_id(request.saml_session, session_info['name_id'])
+        logger.debug("User %s authenticated via SSO.", user)
+        logger.debug('Sending the post_authenticated signal')
+
+        # post_authenticated.send_robust(sender=user, session_info=session_info)
+        # https://github.com/knaperek/djangosaml2/issues/117
+        post_authenticated.send_robust(sender=user.__class__,
+                                       instance=user,
+                                       session_info=session_info,
+                                       request=request)
+        self.customize_session(user, session_info)
+
+        relay_state = self.build_relay_state()
+        custom_redirect_url = self.custom_redirect(user, relay_state, session_info)
+        if custom_redirect_url:
+            return HttpResponseRedirect(custom_redirect_url)
+        relay_state = validate_referral_url(request, relay_state)
+        logger.debug('Redirecting to the RelayState: %s', relay_state)
+        return HttpResponseRedirect(relay_state)
+
+    def build_relay_state(self):
+        """
+        The relay state is a URL used to redirect the user to the view where they came from.
+        """
+        default_relay_state = get_custom_setting('ACS_DEFAULT_REDIRECT_URL',
+                                                 settings.LOGIN_REDIRECT_URL)
+        relay_state = self.request.POST.get('RelayState', '/')
+        relay_state = self.customize_relay_state(relay_state)
+        if not relay_state:
+            logger.warning('The RelayState parameter exists but is empty')
+            relay_state = default_relay_state
+        return relay_state
+
+    def customize_session(self, user, session_info):
+        """
+        Subclasses can use this for customized functionality around user sessions.
+        """
+
+    def customize_relay_state(self, relay_state):
+        """
+        Subclasses may override this method to implement custom logic for relay state.
+        """
+        return relay_state
+
+    def custom_redirect(self, user, relay_state, session_info):
+        """
+        Subclasses may override this method to implement custom logic for redirect.
+
+        For example, some sites may require user registration if the user has not
+        yet been provisioned.
+        """
+        return None
 
-    # redirect the user to the view where he came from
-    default_relay_state = get_custom_setting('ACS_DEFAULT_REDIRECT_URL',
-                                             settings.LOGIN_REDIRECT_URL)
-    relay_state = request.POST.get('RelayState', default_relay_state)
-    if not relay_state:
-        logger.warning('The RelayState parameter exists but is empty')
-        relay_state = default_relay_state
-    if not is_safe_url_compat(url=relay_state, allowed_hosts={request.get_host()}):
-        relay_state = settings.LOGIN_REDIRECT_URL
-    logger.debug('Redirecting to the RelayState: %s', relay_state)
-    return HttpResponseRedirect(relay_state)
 
 
 @login_required
@@ -367,12 +416,12 @@ def echo_attributes(request,
                     config_loader_path=None,
                     template='djangosaml2/echo_attributes.html'):
     """Example view that echo the SAML attributes of an user"""
-    state = StateCache(request.session)
+    state = StateCache(request.saml_session)
     conf = get_config(config_loader_path, request)
 
     client = Saml2Client(conf, state_cache=state,
-                         identity_cache=IdentityCache(request.session))
-    subject_id = _get_subject_id(request.session)
+                         identity_cache=IdentityCache(request.saml_session))
+    subject_id = _get_subject_id(request.saml_session)
     try:
         identity = client.users.get_identity(subject_id,
                                              check_not_on_or_after=False)
@@ -389,19 +438,26 @@ def logout(request, config_loader_path=None):
     This view initiates the SAML2 Logout request
     using the pysaml2 library to create the LogoutRequest.
     """
-    state = StateCache(request.session)
+    state = StateCache(request.saml_session)
     conf = get_config(config_loader_path, request)
 
     client = Saml2Client(conf, state_cache=state,
-                         identity_cache=IdentityCache(request.session))
-    subject_id = _get_subject_id(request.session)
+                         identity_cache=IdentityCache(request.saml_session))
+    subject_id = _get_subject_id(request.saml_session)
     if subject_id is None:
         logger.warning(
             'The session does not contain the subject id for user %s',
             request.user)
 
-    result = client.global_logout(subject_id)
+    try:
+        result = client.global_logout(subject_id)
+    except LogoutError as exp:
+        logger.exception('Error Handled - SLO not supported by IDP: {}'.format(exp))
+        auth.logout(request)
+        state.sync()
+        return HttpResponseRedirect(settings.LOGOUT_REDIRECT_URL)
 
+    auth.logout(request)
     state.sync()
 
     if not result:
@@ -455,9 +511,9 @@ def do_logout_service(request, data, binding, config_loader_path=None, next_page
     logger.debug('Logout service started')
     conf = get_config(config_loader_path, request)
 
-    state = StateCache(request.session)
+    state = StateCache(request.saml_session)
     client = Saml2Client(conf, state_cache=state,
-                         identity_cache=IdentityCache(request.session))
+                         identity_cache=IdentityCache(request.saml_session))
 
     if 'SAMLResponse' in data:  # we started the logout
         logger.debug('Receiving a logout response from the IdP')
@@ -467,7 +523,8 @@ def do_logout_service(request, data, binding, config_loader_path=None, next_page
 
     elif 'SAMLRequest' in data:  # logout started by the IdP
         logger.debug('Receiving a logout request from the IdP')
-        subject_id = _get_subject_id(request.session)
+        subject_id = _get_subject_id(request.saml_session)
+
         if subject_id is None:
             logger.warning(
                 'The session does not contain the subject id for user %s. Performing local logout',
@@ -516,7 +573,7 @@ def metadata(request, config_loader_path=None, valid_for=None):
     """
     conf = get_config(config_loader_path, request)
     metadata = entity_descriptor(conf)
-    return HttpResponse(content=text_type(metadata).encode('utf-8'),
+    return HttpResponse(content=str(metadata).encode('utf-8'),
                         content_type="text/xml; charset=utf8")
 
 
